@@ -10,6 +10,7 @@ from django.conf import settings
 from django.contrib.postgres.search import SearchVector
 from django.contrib.postgres.search import SearchVectorField
 from django.contrib.postgres.indexes import GinIndex
+from django.contrib.postgres.fields import ArrayField
 
 from torque import utils
 
@@ -24,6 +25,9 @@ class Collection(models.Model):
     key_field = models.TextField()
     last_updated = models.DateTimeField(auto_now=True)
 
+    def object_data_name(self):
+        return self.object_name + "_data"
+
     def clean_documents(self, config):
         # return a reduced list of documents based on permissions defined in config
         # (WikiConfig instance)
@@ -37,8 +41,14 @@ class Collection(models.Model):
             .all()
         ):
             if value.document.id not in new_documents:
-                new_documents[value.document.id] = {"key": value.document.key}
-            new_documents[value.document.id][value.field.name] = value.to_python()
+                new_documents[value.document.id] = {
+                    "key": value.document.key,
+                    "db_document": value.document,
+                    "fields": {},
+                }
+            new_documents[value.document.id]["fields"][
+                value.field.name
+            ] = value.to_python()
         return new_documents.values()
 
     @classmethod
@@ -134,6 +144,22 @@ class Wiki(models.Model):
     username = models.TextField(null=True)
     password = models.TextField(null=True)
 
+    # Represents the wikis that this wiki is connected to, if it
+    # uses TorqueMultiWikiConfig
+    #
+    # This is a list of strings, not a connection to Collection, because
+    # when we delete the other collection (an option we may take), and then
+    # recreate it, this needs to remain the same.  It is, in fact, a represenation
+    # of the configuration from the Torque extension, which exists independently
+    # of the state of the rest of the system.
+    linked_wiki_keys = ArrayField(models.TextField(), default=list)
+
+    def invalidate_linked_wiki_toc_cache(self):
+        for wiki in Wiki.objects.filter(linked_wiki_keys__contains=[self.wiki_key]):
+            TableOfContentsCache.objects.filter(wiki_config__wiki=wiki).update(
+                dirty=True
+            )
+
 
 class WikiConfig(models.Model):
     collection = models.ForeignKey(
@@ -164,7 +190,7 @@ class WikiConfig(models.Model):
     # the search cache has to be re-indexed, which is not a catastrophic error.
     in_config = models.BooleanField(default=False)
 
-    cache_dirty = models.BooleanField(default=False)
+    search_cache_dirty = models.BooleanField(default=False)
 
     def rebuild_search_index(self):
         from django.conf import settings
@@ -172,65 +198,52 @@ class WikiConfig(models.Model):
         SearchCacheDocument.objects.filter(wiki_config=self).delete()
         sc_documents = []
         for document_dict in self.collection.clean_documents(self):
-            document = Document.objects.get(
-                key=document_dict["key"], collection=self.collection
-            )
+            document = document_dict["db_document"]
             filtered_data = {}
             for filter in settings.TORQUE_FILTERS:
-                filtered_data[filter.name()] = filter.document_value(document_dict)
-
-            if not SearchCacheDocument.objects.filter(
-                document=document, wiki_config=self
-            ).exists():
-                sc_documents.append(
-                    SearchCacheDocument(
-                        document=document,
-                        collection=self.collection,
-                        wiki=self.wiki,
-                        group=self.group,
-                        wiki_config=self,
-                        filtered_data=filtered_data,
-                        data=" ".join(list(map(str, document_dict.values()))),
-                    )
+                filtered_data[filter.name()] = filter.document_value(
+                    document_dict["fields"]
                 )
 
+            sc_documents.append(
+                SearchCacheDocument(
+                    document=document,
+                    collection=self.collection,
+                    wiki=self.wiki,
+                    group=self.group,
+                    wiki_config=self,
+                    filtered_data=filtered_data,
+                    data=" ".join(list(map(str, document_dict["fields"].values()))),
+                )
+            )
+
         SearchCacheDocument.objects.bulk_create(sc_documents)
         SearchCacheDocument.objects.filter(wiki_config=self).update(
             data_vector=SearchVector("data")
         )
 
-    def rebuild_template_cache(self):
-        TemplateCacheDocument.objects.filter(wiki_config=self).delete()
+    def rebuild_template_cache(self, template):
+        TemplateCacheDocument.objects.filter(
+            wiki_config=self, template=template
+        ).delete()
         template_documents = []
-        if Template.objects.filter(
-            wiki=self.wiki, type="CSV", is_default=True
-        ).exists():
-            csv_template = Template.objects.get(
-                wiki=self.wiki, type="CSV", is_default=True
-            )
-            jinja_template = jinja_env.from_string(
-                csv_template.template_file.read().decode("utf-8")
-            )
-            for document_dict in self.collection.clean_documents(self):
-                document = Document.objects.get(
-                    key=document_dict["key"], collection=self.collection
+
+        jinja_template = jinja_env.from_string(template.get_file_contents())
+        for document_dict in self.collection.clean_documents(self):
+            document = document_dict["db_document"]
+
+            template_documents.append(
+                TemplateCacheDocument(
+                    document=document,
+                    wiki_config=self,
+                    template=template,
+                    rendered_text=jinja_template.render(
+                        {self.collection.object_name: document_dict["fields"]}
+                    ),
                 )
+            )
 
-                if not TemplateCacheDocument.objects.filter(
-                    document=document, wiki_config=self, template=csv_template
-                ).exists():
-                    template_documents.append(
-                        TemplateCacheDocument(
-                            document=document,
-                            wiki_config=self,
-                            template=csv_template,
-                            rendered_text=jinja_template.render(
-                                {self.collection.object_name: document_dict}
-                            ),
-                        )
-                    )
-
-            TemplateCacheDocument.objects.bulk_create(template_documents)
+        TemplateCacheDocument.objects.bulk_create(template_documents)
 
 
 class Document(models.Model):
@@ -339,8 +352,11 @@ class Template(models.Model):
     # the tempate cache needs to be reindexed, which is not the worst
     in_config = models.BooleanField(default=False)
 
+    dirty = models.BooleanField(default=False)
+
     def get_file_contents(self):
-        return b"".join(self.template_file.open().readlines()).decode("utf-8")
+        if self.template_file:
+            return b"".join(self.template_file.open().readlines()).decode("utf-8")
 
     class Meta:
         constraints = [
@@ -364,30 +380,48 @@ class TableOfContents(models.Model):
     raw = models.BooleanField(default=False)
 
     def render_to_mwiki(self, wiki_config):
-        collection = wiki_config.collection
-        documents = collection.clean_documents(wiki_config)
-
         data = json.loads(self.json_file)
-        data[collection.name] = {
-            document[collection.key_field]: document for document in documents
-        }
 
-        toc_templates = Template.objects.filter(
-            collection=collection,
+        global_collection_data_name = wiki_config.collection.object_data_name()
+        data[global_collection_data_name] = {}
+        data["toc_lines"] = {}
+
+        global_toc_templates = Template.objects.filter(
+            collection=wiki_config.collection,
             wiki=wiki_config.wiki,
             type="TOC",
         )
-        line_template = toc_templates.get(is_default=True)
+        global_line_template = global_toc_templates.get(is_default=True)
+        global_line_template_contents = global_line_template.get_file_contents()
 
-        line_template_contents = line_template.template_file.read().decode("utf-8")
-        template_contents = self.template.template_file.read().decode("utf-8")
+        def add_to_data(wiki_config):
+            collection = wiki_config.collection
+            documents = collection.clean_documents(wiki_config)
+            data[global_collection_data_name][collection.name] = {
+                document["fields"][collection.key_field]: document["fields"]
+                for document in documents
+            }
+            data["toc_lines"][collection.name] = {
+                document["fields"][collection.key_field]: jinja_env.from_string(
+                    global_line_template_contents
+                ).render({collection.object_name: document["fields"]})
+                for document in documents
+            }
 
-        data["toc_lines"] = {
-            document[collection.key_field]: jinja_env.from_string(
-                line_template_contents
-            ).render({collection.object_name: document})
-            for document in documents
-        }
+        add_to_data(wiki_config)
+
+        for wiki_key in wiki_config.wiki.linked_wiki_keys:
+            try:
+                linked_wiki_config = WikiConfig.objects.get(
+                    group=wiki_config.group, wiki=Wiki.objects.get(wiki_key=wiki_key)
+                )
+                add_to_data(linked_wiki_config)
+            except WikiConfig.DoesNotExist:
+                # There's no group or wiki specified for this configuration, so we can't build
+                # the toc for that.  Which is ok!
+                pass
+
+        template_contents = self.template.get_file_contents()
         return jinja_env.from_string(template_contents).render(data)
 
     class Meta:
