@@ -239,8 +239,13 @@ def edit_record(collection_name, key, group, wiki, field, new_value):
     models.TableOfContentsCache.objects.filter(
         toc__in=collection.tables_of_contents.all()
     ).update(dirty=True)
+    wiki.invalidate_linked_wiki_toc_cache()
     models.SearchCacheDocument.objects.filter(document=document).update(dirty=True)
 
+    # This is overkill, but that's ok.  There's a bit of work to make it so
+    # the individaul template cache documents have a dirty method
+    collection.templates.update(dirty=True)
+
     collection.last_updated = datetime.now
     collection.save()
 
@@ -448,6 +453,7 @@ def reset_config(request, collection_name, wiki_key):
     wiki.password = None
     wiki.script_path = None
     wiki.server = None
+    wiki.linked_wiki_keys = list()
     wiki.save()
 
     models.WikiConfig.objects.filter(
@@ -467,10 +473,12 @@ def reset_config(request, collection_name, wiki_key):
 # all nicely line up with the other config requests
 def set_wiki_config(request, collection_name, wiki_key):
     wiki = models.Wiki.objects.get_or_create(wiki_key=wiki_key)[0]
-    wiki.username = request.POST["username"]
-    wiki.password = request.POST["password"]
-    wiki.script_path = request.POST["script_path"]
-    wiki.server = request.POST["server"]
+    wiki_config = json.loads(request.body)
+    wiki.username = wiki_config["username"]
+    wiki.password = wiki_config["password"]
+    wiki.script_path = wiki_config["script_path"]
+    wiki.server = wiki_config["server"]
+    wiki.linked_wiki_keys = wiki_config["linked_wiki_keys"]
     wiki.save()
 
     return HttpResponse(status=200)
@@ -508,12 +516,12 @@ def set_group_config(request, collection_name, wiki_key):
             )
             config.save()
 
-            for toc in collection.tables_of_contents.all():
-                (cache, created) = models.TableOfContentsCache.objects.update_or_create(
-                    toc=toc, wiki_config=config
-                )
-                cache.dirty = True
-                cache.save()
+        for toc in collection.tables_of_contents.all():
+            (cache, created) = models.TableOfContentsCache.objects.update_or_create(
+                toc=toc, wiki_config=config
+            )
+            cache.dirty = True
+            cache.save()
 
         config.search_cache_sha = permissions_sha
 
@@ -526,7 +534,22 @@ def set_group_config(request, collection_name, wiki_key):
         config.save()
         config.valid_ids.add(*valid_documents)
         config.valid_fields.add(*valid_fields)
-        config.cache_dirty = True
+        config.search_cache_dirty = True
+        collection.templates.update(dirty=True)
+
+        for linked_wiki in models.Wiki.objects.filter(
+            linked_wiki_keys__contains=[wiki_key]
+        ):
+            try:
+                linked_wiki_config = models.WikiConfig.objects.get(
+                    wiki=linked_wiki, group=config.group
+                )
+                models.TableOfContentsCache.objects.filter(
+                    wiki_config=linked_wiki_config
+                ).update(dirty=True)
+            except models.WikiConfig.DoesNotExist:
+                # If there's no proper config, do nothing
+                pass
 
     config.in_config = True
     config.save()
@@ -556,16 +579,19 @@ def set_template_config(request, collection_name, wiki_key):
 
     collection = models.Collection.objects.get(name=collection_name)
     wiki = models.Wiki.objects.get(wiki_key=wiki_key)
-    config = models.Template.objects.get_or_create(
+    template = models.Template.objects.get_or_create(
         collection=collection, wiki=wiki, type=conf_type, name=conf_name
     )[0]
 
-    config.template_file.save(
-        f"{wiki_key}-{conf_name}", ContentFile(new_config["template"])
-    )
-    config.in_config = True
-    config.is_default = default
-    config.save()
+    if template.get_file_contents() != new_config["template"]:
+        template.template_file.save(
+            f"{wiki_key}-{conf_name}", ContentFile(new_config["template"])
+        )
+        template.dirty = True
+
+    template.in_config = True
+    template.is_default = default
+    template.save()
 
     return HttpResponse(status=200)
 
@@ -581,6 +607,7 @@ def upload_collection(request):
             file=f,
         )
     collection.save()
+    collection.templates.update(dirty=True)
 
     # Regenerate search caches in case data has changed.  We assume that the
     # cache is invalid, making uploading a collection be a very expensive operation,
@@ -588,9 +615,12 @@ def upload_collection(request):
     # and failing.
 
     for config in models.WikiConfig.objects.filter(collection=collection):
-        config.cache_dirty = True
+        config.search_cache_dirty = True
         config.save()
 
+    for wiki in models.Wiki.objects.filter(configs__collection=collection):
+        wiki.invalidate_linked_wiki_toc_cache()
+
     return HttpResponse(status=200)
 
 
@@ -641,13 +671,25 @@ def upload_attachment(request):
         collection=collection, name=request.POST["permissions_field"]
     )
     document = collection.documents.get(key=request.POST["object_id"])
-    (attachment, changed) = models.Attachment.objects.update_or_create(
-        collection=collection,
-        name=secure_filename(request.POST["attachment_name"]),
-        document=document,
-        permissions_field=permissions_field,
-    )
-    attachment.permissions_field = permissions_field
+
+    # Can't use update_or_create here because permissions field is not part
+    # of the unique key (collection, name, document), but also not nullable,
+    # so we have to manually do the update or create.
+    try:
+        attachment = models.Attachment.objects.get(
+            collection=collection,
+            name=secure_filename(request.POST["attachment_name"]),
+            document=document,
+        )
+        attachment.permissions_field = permissions_field
+    except models.Attachment.DoesNotExist:
+        attachment = models.Attachment.objects.create(
+            collection=collection,
+            name=secure_filename(request.POST["attachment_name"]),
+            document=document,
+            permissions_field=permissions_field,
+        )
+
     attachment.file = request.FILES["attachment"]
     attachment.save()
 
